#!/usr/bin/env python3
"""Convert a Keras model to frugally-deep format.
"""

import argparse
import base64
import datetime
import hashlib
import json
from typing import Tuple, Union, Mapping, List, Callable, Set, Any, TypeVar

import numpy as np
import numpy.typing  # pylint: disable=unused-import
from keras import backend as K, Layer
from keras.layers import Input, Embedding, CategoryEncoding
from keras.models import Model, load_model
from keras.src import Functional

__author__ = "Tobias Hermann"
__copyright__ = "Copyright 2017, Tobias Hermann"
__license__ = "MIT"
__maintainer__ = "Tobias Hermann, https://github.com/Dobiasd/frugally-deep"
__email__ = "editgym@gmail.com"

NDFloat32Array = np.typing.NDArray[np.float32]
NDUInt32Array = np.typing.NDArray[np.int32]
Shape = Tuple[int, ...]
LayerConfig = Union[None, Mapping[str, Union[float, list[str], list[list[str]]]]]
TensorRepr = Mapping[str, Union[Shape, List[str]]]

TypeT = TypeVar('TypeT')


def as_list(value_or_values: Union[TypeT, List[TypeT]]) -> List[TypeT]:
    """Leave lists untouched, convert non-list types to a singleton list"""
    if isinstance(value_or_values, list):
        return value_or_values
    return [value_or_values]


def keras_shape_to_fdeep_tensor_shape(raw_shape: Shape) -> Shape:
    """Convert a keras shape to an fdeep shape"""
    return raw_shape[1:]


def get_layer_input_shape(layer: Layer) -> Shape:
    """It is stored in a different property depending on the situation."""
    if hasattr(layer, 'batch_shape'):
        return tuple(layer.batch_shape)
    return tuple(layer.input.shape)


def get_layer_input_shape_tensor_shape(layer: Layer) -> Shape:
    """Convert layer input shape to an fdeep shape"""
    return keras_shape_to_fdeep_tensor_shape(get_layer_input_shape(layer))


def show_tensor(tens: NDFloat32Array) -> TensorRepr:
    """Serialize 3-tensor to a dict"""
    return {
        'shape': tens.shape[1:],
        'values': encode_floats(tens.flatten())
    }


def get_model_input_layers(model: Model) -> List[Layer]:
    """Gets the input layers from model.layers in the correct input order."""
    if len(model.inputs) == 1:
        from keras.src.layers.core.input_layer import InputLayer
        input_layers = []
        for layer in model.layers:
            if isinstance(layer, InputLayer):
                input_layers.append(layer)
            return input_layers
    input_layer_names = [model_input.name for model_input in model.inputs]
    model_layers = {layer.name: layer for layer in model.layers}
    return [model_layers[layer_names] for layer_names in input_layer_names]


def measure_predict(model: Model, data_in: List[NDFloat32Array]) -> Tuple[List[NDFloat32Array], float]:
    """Returns output and duration in seconds"""
    start_time = datetime.datetime.now()
    data_out = model.predict(data_in)
    end_time = datetime.datetime.now()
    duration = end_time - start_time
    print('Forward pass took {} s.'.format(duration.total_seconds()))
    return data_out, duration.total_seconds()


def replace_none_with(value: int, shape: Shape) -> Shape:
    """Replace every None with a fixed value."""
    return tuple(list(map(lambda x: x if x is not None else value, shape)))


def get_first_outbound_op(layer: Layer) -> Functional:
    """Determine primary outbound operation"""
    return layer._outbound_nodes[0].operation


def are_embedding_and_category_encoding_layer_positions_ok_for_testing(model: Model) -> bool:
    """
    Test data can only be generated if all Embedding layers
    and CategoryEncoding layers are positioned directly behind the input nodes.
    """

    def embedding_layer_names(model: Model) -> Set[str]:
        layers = model.layers
        result = set()
        for layer in layers:
            if isinstance(layer, Embedding):
                result.add(layer.name)
        layer_type = type(layer).__name__
        if layer_type in ['Model', 'Sequential', 'Functional']:
            result.union(embedding_layer_names(layer))
        return result

    def embedding_layer_names_at_input_nodes(model: Model) -> Set[str]:
        result = set()
        for input_layer in get_model_input_layers(model):
            if input_layer._outbound_nodes and (
                    isinstance(get_first_outbound_op(input_layer), Embedding) or
                    isinstance(get_first_outbound_op(input_layer), CategoryEncoding)):
                result.add(get_first_outbound_op(input_layer).name)
        return result

    return embedding_layer_names(model) == embedding_layer_names_at_input_nodes(model)


def gen_test_data(model: Model) -> Mapping[str, List[TensorRepr]]:
    """Generate data for model verification test."""

    def set_shape_idx_0_to_1_if_none(shape: Shape) -> Shape:
        """Change first element in tuple to 1."""
        if shape[0] is not None:
            return shape
        shape_lst = list(shape)
        shape_lst[0] = 1
        shape = tuple(shape_lst)
        return shape

    def generate_input_data(input_layer: Layer) -> NDFloat32Array:
        """Random data fitting the input shape of a layer."""
        random_fn: Callable[[Shape], Union[NDFloat32Array, NDUInt32Array]]
        if input_layer._outbound_nodes and isinstance(
                get_first_outbound_op(input_layer), Embedding):
            random_fn = lambda size: np.random.randint(
                0, get_first_outbound_op(input_layer).input_dim, size)
        elif input_layer._outbound_nodes and isinstance(
                get_first_outbound_op(input_layer), CategoryEncoding):
            random_fn = lambda size: np.random.randint(
                0, get_first_outbound_op(input_layer).num_tokens, size)
        else:
            random_fn = lambda size: np.random.normal(size=size).astype(np.float32)
        shape = get_layer_input_shape(input_layer)
        data = random_fn(replace_none_with(32, set_shape_idx_0_to_1_if_none(shape))).astype(np.float32)
        return data

    assert are_embedding_and_category_encoding_layer_positions_ok_for_testing(
        model), 'Test data can only be generated if embedding layers are positioned directly after input nodes.'

    data_in: List[NDFloat32Array] = list(map(generate_input_data, get_model_input_layers(model)))

    data_out_test: List[NDFloat32Array]
    warm_up_runs = 3
    test_runs = 5
    for i in range(warm_up_runs):
        if i == 0:
            data_out_test_raw, _ = measure_predict(model, data_in)
            data_out_test = as_list(data_out_test_raw)
        else:
            measure_predict(model, data_in)
    duration_sum = 0.0
    print('Starting performance measurements.')
    for _ in range(test_runs):
        data_out, duration = measure_predict(model, data_in)
        duration_sum = duration_sum + duration
    duration_avg = duration_sum / test_runs
    print('Forward pass took {} s on average.'.format(duration_avg))
    return {
        'inputs': list(map(show_tensor, data_in)),
        'outputs': list(map(show_tensor, data_out_test))
    }


def split_every(size: int, seq: str) -> List[str]:
    """Split a sequence every seq elements."""
    return [seq[pos:pos + size] for pos in range(0, len(seq), size)]


def encode_floats(arr: NDFloat32Array) -> List[str]:
    """Serialize a sequence of floats."""
    return list(split_every(1024, base64.b64encode(arr).decode('ascii')))


def prepare_filter_weights_conv_2d(weights: NDFloat32Array) -> NDFloat32Array:
    """Change dimension order of 2d filter weights to the one used in fdeep"""
    assert len(weights.shape) == 4
    return np.moveaxis(weights, [0, 1, 2, 3], [1, 2, 3, 0]).flatten()


def prepare_filter_weights_slice_conv_2d(weights: NDFloat32Array) -> NDFloat32Array:
    """Change dimension order of 2d filter weights to the one used in fdeep"""
    assert len(weights.shape) == 4
    return np.moveaxis(weights, [0, 1, 2, 3], [1, 2, 0, 3]).flatten()


def prepare_filter_weights_conv_1d(weights: NDFloat32Array) -> NDFloat32Array:
    """Change dimension order of 1d filter weights to the one used in fdeep"""
    assert len(weights.shape) == 3
    return np.moveaxis(weights, [0, 1, 2], [1, 2, 0]).flatten()


def prepare_filter_weights_conv_1d_transpose(weights: NDFloat32Array) -> NDFloat32Array:
    """Change dimension order of 1d filter weights to the one used in fdeep"""
    assert len(weights.shape) == 3
    return np.moveaxis(weights, [0, 1, 2], [1, 0, 2]).flatten()


def prepare_filter_weights_conv_2d_transpose(weights: NDFloat32Array) -> NDFloat32Array:
    """Change dimension order of 2d filter weights to the one used in fdeep"""
    assert len(weights.shape) == 4
    return np.moveaxis(weights, [0, 1, 2, 3], [1, 2, 0, 3]).flatten()


def show_conv_1d_layer(layer: Layer) -> Mapping[str, list[str]]:
    """Serialize Conv1D layer to dict"""
    weights = layer.get_weights()
    assert len(weights) == 1 or len(weights) == 2
    assert len(weights[0].shape) == 3
    weights_flat = prepare_filter_weights_conv_1d(weights[0])
    assert layer.padding in ['valid', 'same', 'causal']
    assert layer.groups == 1
    assert len(get_layer_input_shape(layer)) == 3
    assert get_layer_input_shape(layer)[0] in {None, 1}
    result = {
        'weights': encode_floats(weights_flat)
    }
    if len(weights) == 2:
        bias = weights[1]
        result['bias'] = encode_floats(bias)
    return result


def show_conv_2d_layer(layer: Layer) -> Mapping[str, list[str]]:
    """Serialize Conv2D layer to dict"""
    weights = layer.get_weights()
    assert len(weights) == 1 or len(weights) == 2
    assert len(weights[0].shape) == 4
    weights_flat = prepare_filter_weights_conv_2d(weights[0])
    assert layer.padding in ['valid', 'same']
    assert layer.groups == 1
    assert len(get_layer_input_shape(layer)) == 4
    assert get_layer_input_shape(layer)[0] in {None, 1}
    result = {
        'weights': encode_floats(weights_flat)
    }
    if len(weights) == 2:
        bias = weights[1]
        result['bias'] = encode_floats(bias)
    return result


def show_separable_conv_2d_layer(layer: Layer) -> Mapping[str, list[str]]:
    """Serialize SeparableConv2D layer to dict"""
    weights = layer.get_weights()
    assert layer.depth_multiplier == 1
    assert len(weights) == 2 or len(weights) == 3
    assert len(weights[0].shape) == 4
    assert len(weights[1].shape) == 4

    # probably incorrect for depth_multiplier > 1?
    slice_weights = prepare_filter_weights_slice_conv_2d(weights[0])
    stack_weights = prepare_filter_weights_conv_2d(weights[1])

    assert layer.padding in ['valid', 'same']
    assert len(get_layer_input_shape(layer)) == 4
    assert get_layer_input_shape(layer)[0] in {None, 1}
    result = {
        'slice_weights': encode_floats(slice_weights),
        'stack_weights': encode_floats(stack_weights),
    }
    if len(weights) == 3:
        bias = weights[2]
        result['bias'] = encode_floats(bias)
    return result


def show_depthwise_conv_2d_layer(layer: Layer) -> Mapping[str, list[str]]:
    """Serialize DepthwiseConv2D layer to dict"""
    weights = layer.get_weights()
    assert layer.depth_multiplier == 1
    assert len(weights) in [1, 2]
    assert len(weights[0].shape) == 4

    # probably incorrect for depth_multiplier > 1?
    slice_weights = prepare_filter_weights_slice_conv_2d(weights[0])

    assert layer.padding in ['valid', 'same']
    assert len(get_layer_input_shape(layer)) == 4
    assert get_layer_input_shape(layer)[0] in {None, 1}
    result = {
        'slice_weights': encode_floats(slice_weights),
    }
    if len(weights) == 2:
        bias = weights[1]
        result['bias'] = encode_floats(bias)
    return result


def show_conv_1d_transpose_layer(layer: Layer) -> Mapping[str, list[str]]:
    """Serialize Conv1D transpose layer to dict"""
    weights = layer.get_weights()
    assert len(weights) == 1 or len(weights) == 2
    assert len(weights[0].shape) == 3
    weights_flat = prepare_filter_weights_conv_1d_transpose(weights[0])
    assert layer.padding in ['valid', 'same', 'causal']
    assert layer.strides[0] <= layer.kernel_size[0]
    assert len(get_layer_input_shape(layer)) == 3
    assert get_layer_input_shape(layer)[0] in {None, 1}
    result = {
        'weights': encode_floats(weights_flat)
    }
    if len(weights) == 2:
        bias = weights[1]
        result['bias'] = encode_floats(bias)
    return result


def show_conv_2d_transpose_layer(layer: Layer) -> Mapping[str, list[str]]:
    """Serialize Conv2D transpose layer to dict"""
    weights = layer.get_weights()
    assert len(weights) == 1 or len(weights) == 2
    assert len(weights[0].shape) == 4
    weights_flat = prepare_filter_weights_conv_2d_transpose(weights[0])
    assert layer.padding in ['valid', 'same']
    assert layer.strides[0] <= layer.kernel_size[0]
    assert layer.strides[1] <= layer.kernel_size[1]

    assert sum([
        layer.dilation_rate[0] == layer.dilation_rate[1],
        layer.strides[0] == layer.strides[1],
        layer.kernel_size[0] == layer.kernel_size[1]]) >= 2
    assert len(get_layer_input_shape(layer)) == 4
    assert get_layer_input_shape(layer)[0] in {None, 1}
    result = {
        'weights': encode_floats(weights_flat)
    }
    if len(weights) == 2:
        bias = weights[1]
        result['bias'] = encode_floats(bias)
    return result


def show_batch_normalization_layer(layer: Layer) -> Mapping[str, list[str]]:
    """Serialize batch normalization layer to dict"""
    moving_mean = layer.moving_mean.numpy()
    moving_variance = layer.moving_variance.numpy()
    result = {}
    result['moving_mean'] = encode_floats(moving_mean)
    result['moving_variance'] = encode_floats(moving_variance)
    if layer.center:
        beta = layer.beta.numpy()
        result['beta'] = encode_floats(beta)
    if layer.scale:
        gamma = layer.gamma.numpy()
        result['gamma'] = encode_floats(gamma)
    return result


def show_layer_normalization_layer(layer: Layer) -> Mapping[str, list[str]]:
    """Serialize layer normalization layer to dict"""
    result = {}
    if layer.center:
        beta = layer.beta.numpy()
        result['beta'] = encode_floats(beta)
    if layer.scale:
        gamma = layer.gamma.numpy()
        result['gamma'] = encode_floats(gamma)
    return result


def show_dense_layer(layer: Layer) -> Mapping[str, list[str]]:
    """Serialize dense layer to dict"""
    weights = layer.get_weights()
    assert len(weights) == 1 or len(weights) == 2
    assert len(weights[0].shape) == 2
    weights_flat = weights[0].flatten()
    result = {
        'weights': encode_floats(weights_flat)
    }
    if len(weights) == 2:
        bias = weights[1]
        result['bias'] = encode_floats(bias)
    return result


def show_dot_layer(layer: Layer) -> None:
    """Check valid configuration of Dot layer"""
    assert len(get_layer_input_shape(layer)) == 2
    assert isinstance(layer.axes, int) or (isinstance(layer.axes, list) and len(layer.axes) == 2)
    assert layer.input.shape[0][0] is None
    assert layer.input.shape[1][0] is None
    assert len(layer.output_shape) <= 5


def show_prelu_layer(layer: Layer) -> Mapping[str, list[str]]:
    """Serialize prelu layer to dict"""
    weights = layer.get_weights()
    assert len(weights) == 1
    weights_flat = weights[0].flatten()
    result = {
        'alpha': encode_floats(weights_flat)
    }
    return result


def show_embedding_layer(layer: Layer) -> Mapping[str, list[str]]:
    """Serialize Embedding layer to dict"""
    weights = layer.get_weights()
    assert len(weights) == 1
    result = {
        'weights': encode_floats(weights[0])
    }
    return result


def show_input_layer(layer: Layer) -> None:
    """Serialize input layer to dict"""
    assert not layer.sparse


def show_softmax_layer(layer: Layer) -> None:
    """Serialize softmax layer to dict"""
    assert layer.axis == -1


def show_normalization_layer(layer: Layer) -> Mapping[str, list[str]]:
    """Serialize normalization layer to dict"""
    assert len(layer.axis) <= 1, 'Multiple normalization axes are not supported'
    if len(layer.axis) == 1:
        assert layer.axis[0] in (-1, 1, 2, 3, 4, 5), 'Invalid axis for Normalization layer.'
    return {
        'mean': encode_floats(layer.mean),
        'variance': encode_floats(layer.variance)
    }


def show_upsampling2d_layer(layer: Layer) -> None:
    """Serialize UpSampling2D layer to dict"""
    assert layer.interpolation in ['nearest', 'bilinear']


def show_resizing_layer(layer: Layer) -> None:
    """Serialize Resizing layer to dict"""
    assert layer.interpolation in ['nearest', 'bilinear', 'area']


def show_rescaling_layer(layer: Layer) -> None:
    """Serialize Rescaling layer to dict"""
    assert isinstance(layer.scale, float)


def show_category_encoding_layer(layer: Layer) -> None:
    """Serialize CategoryEncoding layer to dict"""
    assert layer.output_mode in ['multi_hot', 'count', 'one_hot']


def show_attention_layer(layer: Layer) -> Mapping[str, float]:
    """Serialize Attention layer to dict"""
    assert layer.score_mode in ['dot', 'concat']
    data = {}
    if layer.scale is not None:
        data['scale'] = float(layer.scale.numpy())
    if layer.score_mode == 'concat':
        data['concat_score_weight'] = float(layer.concat_score_weight.numpy())
    return data


def show_additive_attention_layer(layer: Layer) -> Mapping[str, List[str]]:
    """Serialize AdditiveAttention layer to dict"""
    data = {}
    if layer.scale is not None:
        data['scale'] = encode_floats(layer.scale.numpy())
    return data


def show_multi_head_attention_layer(layer: Layer) -> Mapping[str, List[list[str]]]:
    """Serialize MultiHeadAttention layer to dict"""
    assert layer._output_shape is None
    assert layer._attention_axes == (1,), 'MultiHeadAttention supported only with attention_axes=None'
    return {
        'weight_shapes': list(map(lambda w: list(w.shape), layer.weights)),
        'weights': list(map(lambda w: encode_floats(w.numpy()), layer.weights)),
    }


def get_layer_functions_dict() -> Mapping[str, Callable[[Layer], LayerConfig]]:
    return {
        'Conv1D': show_conv_1d_layer,
        'Conv2D': show_conv_2d_layer,
        'Conv1DTranspose': show_conv_1d_transpose_layer,
        'Conv2DTranspose': show_conv_2d_transpose_layer,
        'SeparableConv2D': show_separable_conv_2d_layer,
        'DepthwiseConv2D': show_depthwise_conv_2d_layer,
        'BatchNormalization': show_batch_normalization_layer,
        'Dense': show_dense_layer,
        'Dot': show_dot_layer,
        'PReLU': show_prelu_layer,
        'Embedding': show_embedding_layer,
        'LayerNormalization': show_layer_normalization_layer,
        'TimeDistributed': show_time_distributed_layer,
        'Input': show_input_layer,
        'Softmax': show_softmax_layer,
        'Normalization': show_normalization_layer,
        'UpSampling2D': show_upsampling2d_layer,
        'Resizing': show_resizing_layer,
        'Rescaling': show_rescaling_layer,
        'CategoryEncoding': show_category_encoding_layer,
        'Attention': show_attention_layer,
        'AdditiveAttention': show_additive_attention_layer,
        'MultiHeadAttention': show_multi_head_attention_layer,
    }


def show_time_distributed_layer(layer: Layer) -> Union[None, LayerConfig]:
    show_layer_functions = get_layer_functions_dict()
    config = layer.get_config()
    class_name = str(config['layer']['class_name'])

    if show_layer_functions and class_name in show_layer_functions:
        input_shape_new: Shape
        if len(get_layer_input_shape(layer)) == 3:
            input_shape_new = (get_layer_input_shape(layer)[0], get_layer_input_shape(layer)[2])
        elif len(get_layer_input_shape(layer)) == 4:
            input_shape_new = (
                get_layer_input_shape(layer)[0], get_layer_input_shape(layer)[2], get_layer_input_shape(layer)[3])
        elif len(get_layer_input_shape(layer)) == 5:
            input_shape_new = (
                get_layer_input_shape(layer)[0], get_layer_input_shape(layer)[2], get_layer_input_shape(layer)[3],
                get_layer_input_shape(layer)[4])
        elif len(get_layer_input_shape(layer)) == 6:
            input_shape_new = (
                get_layer_input_shape(layer)[0], get_layer_input_shape(layer)[2], get_layer_input_shape(layer)[3],
                get_layer_input_shape(layer)[4],
                get_layer_input_shape(layer)[5])
        else:
            raise Exception('Wrong input shape')

        layer_function = show_layer_functions[class_name]
        attributes = dir(layer.layer)

        class CopiedLayer:
            pass

        copied_layer = CopiedLayer()

        for attr in attributes:
            try:
                if attr not in ['batch_shape', '__class__']:
                    setattr(copied_layer, attr, getattr(layer.layer, attr))
            except Exception:
                continue

        setattr(copied_layer, 'batch_shape', input_shape_new)
        setattr(copied_layer, 'output_shape', layer.output.shape)

        return layer_function(copied_layer)
    else:
        return None


def get_dict_keys(d: Mapping[str, LayerConfig]) -> list[str]:
    """Return keys of a dictionary"""
    return [key for key in d]


def merge_two_disjunct_dicts(x: Mapping[str, LayerConfig], y: Mapping[str, LayerConfig]) -> Mapping[str, LayerConfig]:
    """Given two dicts, merge them into a new dict as a shallow copy.
    No Key is allowed to be present in both dictionaries.
    """
    assert set(get_dict_keys(x)).isdisjoint(get_dict_keys(y))
    assert isinstance(x, dict) and isinstance(y, dict)
    z = x.copy()
    z.update(y)
    return z


def is_ascii(some_string: str) -> bool:
    """Check if a string only contains ascii characters"""
    try:
        some_string.encode('ascii')
    except UnicodeEncodeError:
        return False
    else:
        return True


def get_layer_weights(layer: Layer, name: str) -> Mapping[str, LayerConfig]:
    """Serialize all weights of a single normal layer"""
    result: dict[str, LayerConfig] = {}
    layer_type = type(layer).__name__
    if hasattr(layer, 'data_format'):
        assert layer.data_format == 'channels_last'

    show_func = get_layer_functions_dict().get(layer_type, None)
    shown_layer = None
    if show_func:
        shown_layer = show_func(layer)
    if shown_layer:
        result[name] = shown_layer
    if show_func and layer_type == 'TimeDistributed':
        result[name] = {'td_input_len': encode_floats(
            np.array([len(get_layer_input_shape(layer)) - 1], dtype=np.float32)),
            'td_output_len': encode_floats(np.array([len(layer.output.shape) - 1], dtype=np.float32))}
    return result


def get_all_weights(model: Model, prefix: str) -> Mapping[str, LayerConfig]:
    """Serialize all weights of the models layers"""
    result: dict[str, LayerConfig] = {}
    layers = model.layers
    assert K.image_data_format() == 'channels_last'
    for layer in layers:
        layer_type = type(layer).__name__
        for node in layer._inbound_nodes:
            if 'training' in node.arguments.kwargs:
                is_layer_with_accidental_training_flag = layer_type in ('CenterCrop', 'Resizing')
                has_training = node.arguments.kwargs['training'] is True
                assert not has_training or is_layer_with_accidental_training_flag, \
                    'training=true is not supported, see https://github.com/Dobiasd/frugally-deep/issues/284'

        name = prefix + layer.name
        assert is_ascii(name)
        if name in result:
            raise ValueError('duplicate layer name ' + name)
        if layer_type in ['Model', 'Sequential', 'Functional']:
            result = dict(merge_two_disjunct_dicts(result, get_all_weights(layer, name + '_')))
        elif layer_type in ['TimeDistributed'] and type(layer.layer).__name__ in ['Model', 'Sequential', 'Functional']:
            inner_layer = layer.layer
            result = dict(merge_two_disjunct_dicts(result, get_layer_weights(layer, name)))
            result = dict(merge_two_disjunct_dicts(result, get_all_weights(inner_layer, name + '_')))
        else:
            result = dict(merge_two_disjunct_dicts(result, get_layer_weights(layer, name)))
    return result


def get_model_name(model: Model) -> str:
    """Return .name or ._name"""
    if hasattr(model, 'name'):
        return str(model.name)
    return str(model._name)


def convert_sequential_to_model(model: Model) -> Model:
    """Convert a sequential model to the underlying functional format"""
    if type(model).__name__ in ['Sequential']:
        name = get_model_name(model)
        inbound_nodes = model._inbound_nodes
        input_layer = Input(batch_shape=get_layer_input_shape(model.layers[0]))
        prev_layer = input_layer
        for layer in model.layers:
            layer._inbound_nodes = []
            prev_layer = layer(prev_layer)
        funcmodel = Model([input_layer], [prev_layer], name=name)
        model = funcmodel
        model._inbound_nodes = inbound_nodes
    if type(model).__name__ == 'TimeDistributed':
        model.layer = convert_sequential_to_model(model.layer)
    if type(model).__name__ in ['Model', 'Functional']:
        for i in range(len(model.layers)):
            new_layer = convert_sequential_to_model(model.layers[i])
            if new_layer == model.layers[i]:
                continue
            # https://stackoverflow.com/questions/78297541/how-to-replace-a-model-layer-using-tensorflow-2-16
            model._operations[i] = new_layer
            assert model.layers[i] == new_layer
    return model


def get_shapes(tensors: List[Mapping[str, Shape]]) -> List[Shape]:
    """Return shapes of a list of tensors"""
    return [t['shape'] for t in tensors]


def calculate_hash(model: Model) -> str:
    layers = model.layers
    hash_m = hashlib.sha256()
    for layer in layers:
        for weights in layer.get_weights():
            if isinstance(weights, np.ndarray):
                hash_m.update(weights.tobytes())
        hash_m.update(layer.name.encode('ascii'))
    return hash_m.hexdigest()


def model_to_fdeep_json(model: Model, no_tests: bool = False) -> Mapping[str, Any]:
    """Convert any Keras model to the frugally-deep model format."""

    # Force creation of underlying functional model.
    # see: https://github.com/fchollet/keras/issues/8136
    # Loss and optimizer type do not matter, since we do not train the model.
    model.compile(loss='mse', optimizer='sgd')

    model = convert_sequential_to_model(model)

    test_data = None if no_tests else gen_test_data(model)

    json_output = {}
    print('Converting model architecture.')
    json_output['architecture'] = json.loads(model.to_json())
    json_output['image_data_format'] = K.image_data_format()
    json_output['input_shapes'] = list(map(get_layer_input_shape_tensor_shape, get_model_input_layers(model)))
    json_output['output_shapes'] = list(map(keras_shape_to_fdeep_tensor_shape, as_list(model.output_shape)))

    if test_data:
        json_output['tests'] = [test_data]

    print('Converting model weights.')
    json_output['trainable_params'] = get_all_weights(model, '')
    print('Done converting model weights.')

    print('Calculating model hash.')
    json_output['hash'] = calculate_hash(model)
    print('Model conversion finished.')

    return json_output


def assert_model_type(model: Model) -> None:
    import keras
    assert type(model) in [keras.src.models.sequential.Sequential, keras.src.models.functional.Functional]


def convert(in_path: str, out_path: str, no_tests: bool = False) -> None:
    """Convert any (h5-)stored Keras model to the frugally-deep model format."""
    print('loading {}'.format(in_path))
    model = load_model(in_path, compile=False)
    json_output = model_to_fdeep_json(model, no_tests)
    print('writing {}'.format(out_path))

    with open(out_path, 'w') as f:
        json.dump(json_output, f, allow_nan=False, separators=(',', ':'))


def main() -> None:
    """Parse command line and convert model."""

    parser = argparse.ArgumentParser(
        prog='frugally-deep model converter',
        description='Converts models from Keras\' .keras format to frugally-deep\'s .json format.')
    parser.add_argument('input_path', type=str)
    parser.add_argument('output_path', type=str)
    parser.add_argument('--no-tests', action='store_true')
    args = parser.parse_args()

    convert(args.input_path, args.output_path, args.no_tests)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
"""Save application models mentioned in Keras documentation
"""

import keras
from keras.models import Model

import convert_model

__author__ = "Tobias Hermann"
__copyright__ = "Copyright 2017, Tobias Hermann"
__license__ = "MIT"
__maintainer__ = "Tobias Hermann, https://github.com/Dobiasd/frugally-deep"
__email__ = "editgym@gmail.com"


def save_model(file_name_base: str, model: Model) -> None:
    """Save and convert Keras model"""
    keras_file = f'{file_name_base}.keras'
    fdeep_file = f'{file_name_base}.json'
    print(f'Saving {keras_file}')
    model.save(keras_file)
    print(f'Converting {keras_file} to {fdeep_file}.')
    convert_model.convert(keras_file, fdeep_file)
    print(f'Conversion of model {keras_file} to {fdeep_file} done.')


def main() -> None:
    """Save famous example models in Keras-h5 and fdeep-json format."""
    print('Saving application examples')
    # save_model('convnextbase', keras.applications.convnext.ConvNeXtBase())  # custom object LayerScale
    # save_model('convnextlarge', keras.applications.convnext.ConvNeXtLarge())  # custom object LayerScale
    # save_model('convnextsmall', keras.applications.convnext.ConvNeXtSmall())  # custom object LayerScale
    # save_model('convnexttiny', keras.applications.convnext.ConvNeXtTiny())  # custom object LayerScale
    # save_model('convnextxlarge', keras.applications.convnext.ConvNeXtXLarge())  # custom object LayerScale
    save_model('densenet121', keras.applications.densenet.DenseNet121())
    save_model('densenet169', keras.applications.densenet.DenseNet169())
    save_model('densenet201', keras.applications.densenet.DenseNet201())
    save_model('efficientnetb0', keras.applications.efficientnet.EfficientNetB0(weights=None))
    save_model('efficientnetb1', keras.applications.efficientnet.EfficientNetB1(weights=None))
    save_model('efficientnetb2', keras.applications.efficientnet.EfficientNetB2(weights=None))
    save_model('efficientnetb3', keras.applications.efficientnet.EfficientNetB3(weights=None))
    save_model('efficientnetb4', keras.applications.efficientnet.EfficientNetB4(weights=None))
    save_model('efficientnetb5', keras.applications.efficientnet.EfficientNetB5(weights=None))
    save_model('efficientnetb6', keras.applications.efficientnet.EfficientNetB6(weights=None))
    save_model('efficientnetb7', keras.applications.efficientnet.EfficientNetB7(weights=None))
    save_model('efficientnetv2b0', keras.applications.efficientnet_v2.EfficientNetV2B0())
    save_model('efficientnetv2b1', keras.applications.efficientnet_v2.EfficientNetV2B1())
    save_model('efficientnetv2b2', keras.applications.efficientnet_v2.EfficientNetV2B2())
    save_model('efficientnetv2b3', keras.applications.efficientnet_v2.EfficientNetV2B3())
    save_model('efficientnetv2l', keras.applications.efficientnet_v2.EfficientNetV2L())
    save_model('efficientnetv2m', keras.applications.efficientnet_v2.EfficientNetV2M())
    save_model('efficientnetv2s', keras.applications.efficientnet_v2.EfficientNetV2S())
    # save_model('inceptionresnetv2', keras.applications.inception_resnet_v2.InceptionResNetV2(input_shape=(299, 299, 3)))  # CustomScaleLayer
    save_model('inceptionv3', keras.applications.inception_v3.InceptionV3(input_shape=(299, 299, 3)))
    save_model('mobilenet', keras.applications.mobilenet.MobileNet())
    save_model('mobilenetv2', keras.applications.mobilenet_v2.MobileNetV2())
    save_model('nasnetlarge', keras.applications.nasnet.NASNetLarge(input_shape=(331, 331, 3)))
    save_model('nasnetmobile', keras.applications.nasnet.NASNetMobile(input_shape=(224, 224, 3)))
    save_model('resnet101', keras.applications.ResNet101())
    save_model('resnet101v2', keras.applications.ResNet101V2())
    save_model('resnet152', keras.applications.ResNet152())
    save_model('resnet152v2', keras.applications.ResNet152V2())
    save_model('resnet50', keras.applications.ResNet50())
    save_model('resnet50v2', keras.applications.ResNet50V2())
    save_model('vgg16', keras.applications.vgg16.VGG16())
    save_model('vgg19', keras.applications.vgg19.VGG19())
    save_model('xception', keras.applications.xception.Xception(input_shape=(299, 299, 3)))


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""Generate a test model for frugally-deep.
"""

import sys
from typing import Tuple, List, Union

import numpy as np
from keras import activations
from keras.layers import ActivityRegularization
from keras.layers import AdditiveAttention
from keras.layers import Attention
from keras.layers import BatchNormalization, Concatenate, LayerNormalization, UnitNormalization
from keras.layers import CategoryEncoding, Embedding
from keras.layers import Conv1D, ZeroPadding1D, Cropping1D
from keras.layers import Conv2D, ZeroPadding2D, Cropping2D, CenterCrop
from keras.layers import GlobalAveragePooling1D, GlobalMaxPooling1D
from keras.layers import GlobalAveragePooling2D, GlobalMaxPooling2D
from keras.layers import GlobalAveragePooling3D, GlobalMaxPooling3D
from keras.layers import Identity, Conv2DTranspose, Conv1DTranspose
from keras.layers import Input, Dense, Dropout, Flatten, Activation
from keras.layers import LeakyReLU, ELU, PReLU, ReLU, Softmax
from keras.layers import MaxPooling1D, AveragePooling1D, UpSampling1D
from keras.layers import MaxPooling2D, AveragePooling2D, UpSampling2D
from keras.layers import MaxPooling3D, AveragePooling3D
from keras.layers import MultiHeadAttention
from keras.layers import Multiply, Add, Subtract, Average, Maximum, Minimum
from keras.layers import Normalization, Rescaling, Resizing
from keras.layers import Permute, Reshape, RepeatVector
from keras.layers import SeparableConv2D, DepthwiseConv2D
from keras.layers import ZeroPadding3D, Cropping3D
from keras.models import Model, load_model, Sequential

__author__ = "Tobias Hermann"
__copyright__ = "Copyright 2017, Tobias Hermann"
__license__ = "MIT"
__maintainer__ = "Tobias Hermann, https://github.com/Dobiasd/frugally-deep"
__email__ = "editgym@gmail.com"

NDFloat32Array = np.typing.NDArray[np.float32]
NDUInt32Array = np.typing.NDArray[np.int32]
Shape = Tuple[int, ...]
VariableShape = Tuple[Union[None, int], ...]


def replace_none_with(value: int, shape: VariableShape) -> Shape:
    """Replace every None with a fixed value."""
    return tuple(list(map(lambda x: x if x is not None else value, shape)))


def get_shape_for_random_data(data_size: int, shape: Shape) -> Shape:
    """Include size of data to generate into shape."""
    if len(shape) == 5:
        return (data_size, shape[0], shape[1], shape[2], shape[3], shape[4])
    if len(shape) == 4:
        return (data_size, shape[0], shape[1], shape[2], shape[3])
    if len(shape) == 3:
        return (data_size, shape[0], shape[1], shape[2])
    if len(shape) == 2:
        return (data_size, shape[0], shape[1])
    if len(shape) == 1:
        return (data_size, shape[0])
    raise ValueError('can not use shape for random data:', shape)


def generate_random_data(data_size: int, shape: VariableShape) -> NDFloat32Array:
    """Random data for training."""
    return np.random.random(
        size=get_shape_for_random_data(data_size, replace_none_with(42, shape))).astype(np.float32)


def generate_input_data(data_size: int, input_shapes: List[VariableShape]) -> List[NDFloat32Array]:
    """Random input data for training."""
    return [generate_random_data(data_size, input_shape)
            for input_shape in input_shapes]


def generate_integer_random_data(data_size: int, low: int, high: int, shape: Shape) -> NDUInt32Array:
    """Random data for training."""
    return np.random.randint(
        low=low, high=high, size=get_shape_for_random_data(data_size, replace_none_with(42, shape)))


def generate_integer_input_data(data_size: int, low: int, highs: List[int], input_shapes: List[Shape]) -> List[
    NDUInt32Array]:
    """Random input data for training."""
    return [generate_integer_random_data(data_size, low, high, input_shape)
            for high, input_shape in zip(highs, input_shapes)]


def as_list(value_or_values: Union[NDFloat32Array, List[NDFloat32Array]]) -> List[NDFloat32Array]:
    """Leave lists untouched, convert non-list types to a singleton list"""
    if isinstance(value_or_values, list):
        return value_or_values
    return [value_or_values]


def generate_output_data(data_size: int, outputs: List[NDFloat32Array]) -> List[NDFloat32Array]:
    """Random output data for training."""
    return [generate_random_data(data_size, output.shape[1:])
            for output in as_list(outputs)]


def get_test_model_exhaustive() -> Model:
    """Returns a exhaustive test model."""
    input_shapes: List[VariableShape] = [
        (2, 3, 4, 5, 6),  # 0
        (2, 3, 4, 5, 6),
        (7, 8, 9, 10),
        (7, 8, 9, 10),
        (11, 12, 13),
        (11, 12, 13),
        (14, 15),
        (14, 15),
        (16,),
        (16,),
        (2,),  # 10
        (1,),
        (2,),
        (1,),
        (1, 3),
        (1, 4),
        (1, 1, 3),
        (1, 1, 4),
        (1, 1, 1, 3),
        (1, 1, 1, 4),
        (1, 1, 1, 1, 3),  # 20
        (1, 1, 1, 1, 4),
        (26, 28, 3),
        (4, 4, 3),
        (4, 4, 3),
        (4,),
        (2, 3),
        (1,),
        (1,),
        (1,),
        (2, 3),  # 30
        (9, 16, 1),
        (1, 9, 16),
        (6, 1, 1),
        (1, 1, 1, 1, 6),
        (1, 1, 1, 10),
        (1, 1, 13),
        (1, 15),
        (1, 1, 1, 1, 6),
        (1, 1, 1, 5, 1),
        (1, 1, 4, 1, 1),  # 40
        (1, 3, 1, 1, 1),
        (2, 1, 1, 1, 1),
        (1, 1, 4, 1, 6),
        (1, 3, 1, 5, 1),
        (2, 1, 4, 1, 1),  # 45
        (1,),
        (3, 1),
        (6, 5, 4, 3, 2),
        (5, 4),
        (7, 4),  # 50
        (7, 4),
        (2, 4, 6, 8, 10),
    ]

    inputs = [Input(shape=s) for s in input_shapes]

    outputs = []

    outputs.append(Conv1DTranspose(4, 3, padding='valid', use_bias=False)(inputs[6]))
    outputs.append(Conv2DTranspose(4, (5, 3), padding='valid', use_bias=False)(inputs[4]))

    outputs.append(Conv1DTranspose(4, 1, padding='valid', use_bias=False)(inputs[6]))
    outputs.append(Conv1DTranspose(4, 1, padding='same', use_bias=False)(inputs[6]))
    outputs.append(Conv1DTranspose(1, 3, padding='valid', use_bias=False)(inputs[6]))
    outputs.append(Conv1DTranspose(2, 3, padding='same')(inputs[6]))
    outputs.append(Conv1DTranspose(2, 5, padding='same', strides=2)(inputs[6]))
    outputs.append(Conv1DTranspose(2, 5, padding='valid', strides=2)(inputs[6]))
    # Current CPU implementations do not yet support dilation rates larger than 1
    # https://github.com/keras-team/keras/issues/20408
    # https://github.com/keras-team/keras/pull/20737
    #outputs.append(Conv1DTranspose(3, 5, padding='same', dilation_rate=2)(inputs[6]))
    #outputs.append(Conv1DTranspose(3, 5, padding='valid', dilation_rate=2)(inputs[6]))

    outputs.append(Conv2DTranspose(4, (1, 1), padding='valid', use_bias=False)(inputs[4]))
    outputs.append(Conv2DTranspose(4, (1, 1), padding='same', use_bias=False)(inputs[4]))
    outputs.append(Conv2DTranspose(1, (3, 3))(inputs[4]))
    outputs.append(Conv2DTranspose(4, (3, 3), padding='valid', use_bias=False)(inputs[4]))
    outputs.append(Conv2DTranspose(4, (3, 3), padding='same')(inputs[4]))
    outputs.append(Conv2DTranspose(4, (5, 5), padding='same', strides=(2, 3))(inputs[4]))
    outputs.append(Conv2DTranspose(4, (5, 5), padding='valid', strides=(2, 3))(inputs[4]))
    #outputs.append(Conv2DTranspose(4, (5, 5), padding='same', dilation_rate=(2, 3))(inputs[4]))
    #outputs.append(Conv2DTranspose(4, (5, 5), padding='valid', dilation_rate=(2, 3))(inputs[4]))

    outputs.append(Conv1DTranspose(1, 3, padding='valid')(inputs[6]))
    outputs.append(Conv1DTranspose(2, 1, padding='same')(inputs[6]))
    #outputs.append(Conv1DTranspose(3, 4, padding='same', dilation_rate=2)(inputs[6]))

    outputs.append(Conv2DTranspose(4, (3, 3))(inputs[4]))
    outputs.append(Conv2DTranspose(4, (3, 3), use_bias=False)(inputs[4]))
    outputs.append(Conv2DTranspose(4, (2, 4), strides=(2, 2), padding='same')(inputs[4]))
    #outputs.append(Conv2DTranspose(4, (2, 4), padding='same', dilation_rate=(2, 2))(inputs[4]))

    outputs.append(Conv1D(1, 3, padding='valid')(inputs[6]))
    outputs.append(Conv1D(2, 1, padding='same')(inputs[6]))
    outputs.append(Conv1D(2, 1, padding='same', strides=2)(inputs[6]))
    outputs.append(Conv1D(3, 4, padding='causal', dilation_rate=2)(inputs[6]))
    outputs.append(ZeroPadding1D(2)(inputs[6]))
    outputs.append(Cropping1D((2, 3))(inputs[6]))
    outputs.append(MaxPooling1D(2)(inputs[6]))
    outputs.append(MaxPooling1D(2, strides=2, padding='same')(inputs[6]))
    outputs.append(AveragePooling1D(2)(inputs[6]))
    outputs.append(AveragePooling1D(2, strides=2, padding='same')(inputs[6]))
    outputs.append(GlobalMaxPooling1D()(inputs[6]))
    outputs.append(GlobalAveragePooling1D()(inputs[6]))
    outputs.append(GlobalMaxPooling1D(keepdims=True)(inputs[6]))
    outputs.append(GlobalAveragePooling1D(keepdims=True)(inputs[6]))

    outputs.append(Normalization(axis=None, mean=2.1, variance=2.2)(inputs[4]))
    outputs.append(Normalization(axis=None, mean=2.1, variance=2.2, invert=True)(inputs[4]))
    # outputs.append(Normalization(axis=-1, mean=2.1, variance=2.2)(inputs[6]))  # No longer supported in TensorFlow 2.16
    outputs.append(Normalization(axis=-1, mean=2.1, variance=2.2)(inputs[46]))
    outputs.append(Normalization(axis=1, mean=2.1, variance=2.2)(inputs[46]))
    outputs.append(Normalization(axis=-1, mean=2.1, variance=2.2)(inputs[47]))
    # outputs.append(Normalization(axis=1, mean=2.1, variance=2.2)(inputs[47]))  # No longer supported in TensorFlow 2.16
    outputs.append(Normalization(axis=2, mean=2.1, variance=2.2)(inputs[47]))
    for axis in range(1, 6):
        outputs.append(Normalization(axis=axis, mean=4.2, variance=2.3)(inputs[0]))

    outputs.append(Rescaling(23.5, 42.1)(inputs[0]))

    outputs.append(Conv2D(4, (3, 3))(inputs[4]))
    outputs.append(Conv2D(4, (3, 3), use_bias=False, padding='valid')(inputs[4]))
    outputs.append(Conv2D(4, (2, 4), strides=(2, 3), padding='same')(inputs[4]))
    outputs.append(Conv2D(4, (2, 4), padding='same', dilation_rate=(2, 3))(inputs[4]))

    outputs.append(SeparableConv2D(3, (3, 3))(inputs[4]))
    outputs.append(DepthwiseConv2D((3, 3))(inputs[4]))
    outputs.append(DepthwiseConv2D((1, 2))(inputs[4]))

    outputs.append(MaxPooling2D((2, 2))(inputs[4]))
    outputs.append(MaxPooling3D((2, 2, 2))(inputs[2]))
    outputs.append(MaxPooling2D((1, 3), strides=(2, 3), padding='same')(inputs[4]))
    outputs.append(MaxPooling3D((1, 3, 5), strides=(2, 3, 4), padding='same')(inputs[2]))
    outputs.append(AveragePooling2D((2, 2))(inputs[4]))
    outputs.append(AveragePooling3D((2, 2, 2))(inputs[2]))
    outputs.append(AveragePooling2D((1, 3), strides=(2, 3), padding='same')(inputs[4]))
    outputs.append(AveragePooling3D((1, 3, 5), strides=(2, 3, 4), padding='same')(inputs[2]))

    outputs.append(GlobalAveragePooling2D()(inputs[4]))
    outputs.append(GlobalAveragePooling3D()(inputs[2]))
    outputs.append(GlobalMaxPooling2D()(inputs[4]))
    outputs.append(GlobalMaxPooling3D()(inputs[2]))
    outputs.append(GlobalAveragePooling2D(keepdims=True)(inputs[4]))
    outputs.append(GlobalAveragePooling3D(keepdims=True)(inputs[2]))
    outputs.append(GlobalMaxPooling2D(keepdims=True)(inputs[4]))
    outputs.append(GlobalMaxPooling3D(keepdims=True)(inputs[2]))

    outputs.append(CenterCrop(4, 5)(inputs[4]))
    outputs.append(CenterCrop(5, 6)(inputs[4]))
    outputs.append(CenterCrop(19, 53)(inputs[23]))

    outputs.append(UpSampling2D(size=(1, 2), interpolation='nearest')(inputs[4]))
    outputs.append(UpSampling2D(size=(5, 3), interpolation='nearest')(inputs[4]))
    outputs.append(UpSampling2D(size=(1, 2), interpolation='bilinear')(inputs[4]))
    outputs.append(UpSampling2D(size=(5, 3), interpolation='bilinear')(inputs[4]))

    outputs.append(Resizing(4, 5)(inputs[4]))
    outputs.append(Resizing(5, 6)(inputs[4]))
    outputs.append(Resizing(19, 53, interpolation="bilinear")(inputs[23]))
    outputs.append(Resizing(19, 53, interpolation="nearest")(inputs[23]))
    # outputs.append(Resizing(7, 9, interpolation="area")(inputs[22]))  # No longer supported in TensorFlow 2.16
    # outputs.append(Resizing(19, 53, interpolation="area")(inputs[23]))  # No longer supported in TensorFlow 2.16
    outputs.append(Resizing(19, 53, crop_to_aspect_ratio=True)(inputs[23]))

    outputs.append(Permute((3, 4, 1, 5, 2))(inputs[0]))
    outputs.append(Permute((1, 5, 3, 2, 4))(inputs[0]))
    outputs.append(Permute((3, 4, 1, 2))(inputs[2]))
    outputs.append(Permute((2, 1, 3))(inputs[4]))
    outputs.append(Permute((2, 1))(inputs[6]))
    outputs.append(Permute((1,))(inputs[8]))
    outputs.append(Permute((3, 1, 2))(inputs[31]))
    outputs.append(Permute((3, 1, 2))(inputs[32]))

    outputs.append(BatchNormalization(center=False, scale=False)(inputs[11]))
    outputs.append(BatchNormalization()(inputs[11]))
    outputs.append(BatchNormalization()(inputs[10]))
    outputs.append(BatchNormalization()(inputs[14]))
    outputs.append(BatchNormalization()(inputs[26]))
    outputs.append(BatchNormalization()(inputs[23]))
    outputs.append(BatchNormalization()(inputs[0]))
    outputs.append(BatchNormalization(center=False)(inputs[0]))
    outputs.append(BatchNormalization(scale=False)(inputs[0]))
    outputs.append(BatchNormalization(center=False, scale=False)(inputs[0]))
    outputs.append(BatchNormalization()(inputs[0]))
    outputs.append(BatchNormalization(axis=1)(inputs[0]))
    outputs.append(BatchNormalization(axis=2)(inputs[0]))
    outputs.append(BatchNormalization(axis=3)(inputs[0]))
    outputs.append(BatchNormalization(axis=4)(inputs[0]))
    outputs.append(BatchNormalization(axis=5)(inputs[0]))
    outputs.append(BatchNormalization()(inputs[2]))
    outputs.append(BatchNormalization(axis=1)(inputs[2]))
    outputs.append(BatchNormalization(axis=2)(inputs[2]))
    outputs.append(BatchNormalization(axis=3)(inputs[2]))
    outputs.append(BatchNormalization(axis=4)(inputs[2]))
    outputs.append(BatchNormalization()(inputs[4]))
    # todo: check if TensorFlow >= 2.1 supports this
    # outputs.append(BatchNormalization(axis=1)(inputs[4])) # tensorflow.python.framework.errors_impl.InternalError:  The CPU implementation of FusedBatchNorm only supports NHWC tensor format for now.
    outputs.append(BatchNormalization(axis=2)(inputs[4]))
    outputs.append(BatchNormalization(axis=3)(inputs[4]))
    outputs.append(BatchNormalization()(inputs[6]))
    outputs.append(BatchNormalization(axis=1)(inputs[6]))
    outputs.append(BatchNormalization(axis=2)(inputs[6]))
    outputs.append(BatchNormalization()(inputs[8]))
    outputs.append(BatchNormalization(axis=1)(inputs[8]))
    outputs.append(BatchNormalization()(inputs[27]))
    outputs.append(BatchNormalization(axis=1)(inputs[27]))
    outputs.append(BatchNormalization()(inputs[14]))
    outputs.append(BatchNormalization(axis=1)(inputs[14]))
    outputs.append(BatchNormalization(axis=2)(inputs[14]))
    outputs.append(BatchNormalization()(inputs[16]))
    # todo: check if TensorFlow >= 2.1 supports this
    # outputs.append(BatchNormalization(axis=1)(inputs[16])) # tensorflow.python.framework.errors_impl.InternalError:  The CPU implementation of FusedBatchNorm only supports NHWC tensor format for now.
    outputs.append(BatchNormalization(axis=2)(inputs[16]))
    outputs.append(BatchNormalization(axis=3)(inputs[16]))
    outputs.append(BatchNormalization()(inputs[18]))
    outputs.append(BatchNormalization(axis=1)(inputs[18]))
    outputs.append(BatchNormalization(axis=2)(inputs[18]))
    outputs.append(BatchNormalization(axis=3)(inputs[18]))
    outputs.append(BatchNormalization(axis=4)(inputs[18]))
    outputs.append(BatchNormalization()(inputs[20]))
    outputs.append(BatchNormalization(axis=1)(inputs[20]))
    outputs.append(BatchNormalization(axis=2)(inputs[20]))
    outputs.append(BatchNormalization(axis=3)(inputs[20]))
    outputs.append(BatchNormalization(axis=4)(inputs[20]))
    outputs.append(BatchNormalization(axis=5)(inputs[20]))

    outputs.append(LayerNormalization()(inputs[11]))
    outputs.append(LayerNormalization()(inputs[10]))
    outputs.append(LayerNormalization()(inputs[26]))
    outputs.append(LayerNormalization()(inputs[24]))
    outputs.append(LayerNormalization()(inputs[0]))
    outputs.append(LayerNormalization(axis=1)(inputs[0]))
    outputs.append(LayerNormalization(axis=2)(inputs[0]))
    outputs.append(LayerNormalization(axis=3)(inputs[0]))
    outputs.append(LayerNormalization(axis=4)(inputs[0]))
    outputs.append(LayerNormalization(axis=5)(inputs[0]))
    outputs.append(LayerNormalization(axis=[1, 2])(inputs[0]))
    outputs.append(LayerNormalization(axis=[2, 3, 5])(inputs[0]))

    outputs.append(UnitNormalization()(inputs[11]))
    outputs.append(UnitNormalization()(inputs[10]))
    outputs.append(UnitNormalization()(inputs[26]))
    outputs.append(UnitNormalization()(inputs[24]))
    outputs.append(UnitNormalization()(inputs[0]))
    outputs.append(UnitNormalization(axis=1)(inputs[0]))
    outputs.append(UnitNormalization(axis=2)(inputs[0]))
    outputs.append(UnitNormalization(axis=3)(inputs[0]))
    outputs.append(UnitNormalization(axis=4)(inputs[0]))
    outputs.append(UnitNormalization(axis=5)(inputs[0]))
    outputs.append(UnitNormalization(axis=[1, 2])(inputs[0]))
    outputs.append(UnitNormalization(axis=[2, 3, 5])(inputs[0]))

    outputs.append(Dropout(0.5)(inputs[4]))
    outputs.append(ActivityRegularization(0.3, 0.4)(inputs[4]))

    outputs.append(ZeroPadding2D(2)(inputs[4]))
    outputs.append(ZeroPadding2D((2, 3))(inputs[4]))
    outputs.append(ZeroPadding2D(((1, 2), (3, 4)))(inputs[4]))
    outputs.append(Cropping2D(2)(inputs[4]))
    outputs.append(Cropping2D((2, 3))(inputs[4]))
    outputs.append(Cropping2D(((1, 2), (3, 4)))(inputs[4]))

    outputs.append(ZeroPadding3D(2)(inputs[2]))
    outputs.append(ZeroPadding3D((2, 3, 4))(inputs[2]))
    outputs.append(ZeroPadding3D(((1, 2), (3, 4), (5, 6)))(inputs[2]))
    outputs.append(Cropping3D(2)(inputs[2]))
    outputs.append(Cropping3D((2, 3, 4))(inputs[2]))
    outputs.append(Cropping3D(((1, 2), (3, 4), (2, 1)))(inputs[2]))

    outputs.append(Dense(3, use_bias=True)(inputs[0]))
    outputs.append(Dense(3, use_bias=True)(inputs[2]))
    outputs.append(Dense(3, use_bias=True)(inputs[4]))
    outputs.append(Dense(3, use_bias=True)(inputs[6]))
    outputs.append(Dense(3, use_bias=True)(inputs[8]))
    outputs.append(Dense(3, use_bias=True)(inputs[13]))
    outputs.append(Dense(3, use_bias=True)(inputs[14]))
    outputs.append(Dense(4, use_bias=False)(inputs[16]))
    outputs.append(Dense(4, use_bias=False, activation='tanh')(inputs[18]))
    outputs.append(Dense(4, use_bias=False)(inputs[20]))

    outputs.append(Reshape(((2 * 3 * 4 * 5 * 6),))(inputs[0]))
    outputs.append(Reshape((2, 3 * 4 * 5 * 6))(inputs[0]))
    outputs.append(Reshape((2, 3, 4 * 5 * 6))(inputs[0]))
    outputs.append(Reshape((2, 3, 4, 5 * 6))(inputs[0]))
    outputs.append(Reshape((2, 3, 4, 5, 6))(inputs[0]))

    outputs.append(Maximum()([inputs[0], inputs[1]]))
    outputs.append(Maximum()([inputs[2], inputs[3]]))
    outputs.append(Maximum()([inputs[4], inputs[5]]))
    outputs.append(Maximum()([inputs[6], inputs[7]]))
    outputs.append(Maximum()([inputs[8], inputs[9]]))

    outputs.append(Minimum()([inputs[0], inputs[1]]))
    outputs.append(Minimum()([inputs[2], inputs[3]]))
    outputs.append(Minimum()([inputs[4], inputs[5]]))
    outputs.append(Minimum()([inputs[6], inputs[7]]))
    outputs.append(Minimum()([inputs[8], inputs[9]]))

    # No longer works in TensorFlow 2.16, see: https://github.com/tensorflow/tensorflow/issues/65056
    # for normalize in [True, False]:
    # outputs.append(Dot(axes=(1, 1), normalize=normalize)([inputs[8], inputs[9]]))
    # outputs.append(Dot(axes=(1, 1), normalize=normalize)([inputs[0], inputs[10]]))
    # outputs.append(Dot(axes=1, normalize=normalize)([inputs[0], inputs[10]]))
    # outputs.append(Dot(axes=(3, 1), normalize=normalize)([inputs[31], inputs[32]]))
    # outputs.append(Dot(axes=(2, 3), normalize=normalize)([inputs[31], inputs[32]]))
    # outputs.append(Dot(axes=(2, 3), normalize=normalize)([inputs[14], inputs[16]]))
    # outputs.append(Dot(axes=(3, 2), normalize=normalize)([inputs[24], inputs[26]]))

    outputs.append(Reshape((16,))(inputs[8]))
    outputs.append(Reshape((2, 8))(inputs[8]))
    outputs.append(Reshape((2, 2, 4))(inputs[8]))
    outputs.append(Reshape((2, 2, 2, 2))(inputs[8]))
    outputs.append(Reshape((2, 2, 1, 2, 2))(inputs[8]))

    outputs.append(RepeatVector(3)(inputs[8]))

    outputs.append(ReLU()(inputs[0]))

    for axis in [-5, -4, -3, -2, -1, 1, 2, 3, 4, 5]:
        outputs.append(Concatenate(axis=axis)([inputs[0], inputs[1]]))
    for axis in [-4, -3, -2, -1, 1, 2, 3, 4]:
        outputs.append(Concatenate(axis=axis)([inputs[2], inputs[3]]))
    for axis in [-3, -2, -1, 1, 2, 3]:
        outputs.append(Concatenate(axis=axis)([inputs[4], inputs[5]]))
    for axis in [-2, -1, 1, 2]:
        outputs.append(Concatenate(axis=axis)([inputs[6], inputs[7]]))
    for axis in [-1, 1]:
        outputs.append(Concatenate(axis=axis)([inputs[8], inputs[9]]))
    for axis in [-1]:  # [-1, 2] no longer supported in TensorFlow 2.16
        outputs.append(Concatenate(axis=axis)([inputs[14], inputs[15]]))
    for axis in [-1, 3]:
        outputs.append(Concatenate(axis=axis)([inputs[16], inputs[17]]))
    # for axis in [-1, 4]:
    # outputs.append(Concatenate(axis=axis)([inputs[18], inputs[19]]))  # no longer supported in TensorFlow 2.16
    # for axis in [-1, 5]:
    # outputs.append(Concatenate(axis=axis)([inputs[20], inputs[21]]))  # no longer supported in TensorFlow 2.16

    outputs.append(UpSampling1D(size=2)(inputs[6]))
    # outputs.append(UpSampling1D(size=2)(inputs[8])) # ValueError: Input 0 of layer up_sampling1d_1 is incompatible with the layer: expected ndim=3, found ndim=2. Full shape received: [None, 16]

    outputs.append(Multiply()([inputs[10], inputs[11]]))
    outputs.append(Multiply()([inputs[11], inputs[10]]))
    outputs.append(Multiply()([inputs[11], inputs[13]]))
    outputs.append(Multiply()([inputs[10], inputs[11], inputs[12]]))
    outputs.append(Multiply()([inputs[11], inputs[12], inputs[13]]))
    outputs.append(Multiply()([inputs[14], inputs[16], inputs[18], inputs[20]]))
    outputs.append(Multiply()([inputs[14], inputs[16]]))
    outputs.append(Multiply()([inputs[16], inputs[18]]))
    outputs.append(Multiply()([inputs[18], inputs[20]]))
    outputs.append(Multiply()([inputs[30], inputs[33]]))
    outputs.append(Multiply()([inputs[34], inputs[0]]))
    outputs.append(Multiply()([inputs[35], inputs[2]]))
    outputs.append(Multiply()([inputs[36], inputs[4]]))
    outputs.append(Multiply()([inputs[37], inputs[6]]))
    outputs.append(Multiply()([inputs[0], inputs[38]]))
    outputs.append(Multiply()([inputs[0], inputs[39]]))
    outputs.append(Multiply()([inputs[0], inputs[40]]))
    outputs.append(Multiply()([inputs[0], inputs[41]]))
    outputs.append(Multiply()([inputs[0], inputs[42]]))
    outputs.append(Multiply()([inputs[43], inputs[44]]))
    outputs.append(Multiply()([inputs[44], inputs[45]]))

    outputs.append(Attention(use_scale=False, score_mode='dot')([inputs[49], inputs[50]]))
    outputs.append(Attention(use_scale=False, score_mode='dot')([inputs[49], inputs[50], inputs[51]]))
    outputs.append(Attention(use_scale=True, score_mode='dot')([inputs[49], inputs[50]]))
    outputs.append(Attention(use_scale=False, score_mode='concat')([inputs[49], inputs[50]]))

    outputs.append(AdditiveAttention(use_scale=False)([inputs[49], inputs[50]]))
    outputs.append(AdditiveAttention(use_scale=False)([inputs[49], inputs[50], inputs[51]]))
    outputs.append(AdditiveAttention(use_scale=True)([inputs[49], inputs[50]]))
    outputs.append(AdditiveAttention(use_scale=True)([inputs[49], inputs[50], inputs[51]]))

    outputs.append(MultiHeadAttention(
        num_heads=1, key_dim=1, value_dim=None,
        use_bias=False, output_shape=None, attention_axes=None)(inputs[49], inputs[50]))
    outputs.append(MultiHeadAttention(
        num_heads=1, key_dim=1, value_dim=None,
        use_bias=True, output_shape=None, attention_axes=None)(inputs[49], inputs[50]))
    outputs.append(MultiHeadAttention(
        num_heads=1, key_dim=2, value_dim=None,
        use_bias=False, output_shape=None, attention_axes=None)(inputs[49], inputs[50]))
    outputs.append(MultiHeadAttention(
        num_heads=1, key_dim=2, value_dim=None,
        use_bias=True, output_shape=None, attention_axes=None)(inputs[49], inputs[50]))
    outputs.append(MultiHeadAttention(
        num_heads=1, key_dim=1, value_dim=2,
        use_bias=False, output_shape=None, attention_axes=None)(inputs[49], inputs[50]))
    outputs.append(MultiHeadAttention(
        num_heads=1, key_dim=1, value_dim=2,
        use_bias=True, output_shape=None, attention_axes=None)(inputs[49], inputs[50]))
    outputs.append(MultiHeadAttention(
        num_heads=3, key_dim=1, value_dim=None,
        use_bias=False, output_shape=None, attention_axes=None)(inputs[49], inputs[50]))
    outputs.append(MultiHeadAttention(
        num_heads=3, key_dim=1, value_dim=None,
        use_bias=True, output_shape=None, attention_axes=None)(inputs[49], inputs[50]))
    outputs.append(MultiHeadAttention(
        num_heads=1, key_dim=1, value_dim=None,
        use_bias=False, output_shape=None, attention_axes=None)(inputs[49], inputs[50], inputs[51]))
    outputs.append(MultiHeadAttention(
        num_heads=2, key_dim=3, value_dim=5,
        use_bias=False, output_shape=None, attention_axes=None)(inputs[49], inputs[50], inputs[51]))
    outputs.append(MultiHeadAttention(
        num_heads=2, key_dim=3, value_dim=5,
        use_bias=True, output_shape=None, attention_axes=None)(inputs[49], inputs[50], inputs[51]))

    shared_conv = Conv2D(1, (1, 1),
                         padding='valid', name='shared_conv', activation='relu')

    up_scale_2 = UpSampling2D((2, 2))
    x1 = shared_conv(up_scale_2(inputs[23]))  # (1, 8, 8)
    x2 = shared_conv(up_scale_2(inputs[24]))  # (1, 8, 8)
    x3 = Conv2D(1, (1, 1), padding='valid')(up_scale_2(inputs[24]))  # (1, 8, 8)
    x = Concatenate()([x1, x2, x3])  # (3, 8, 8)
    outputs.append(x)

    x = Conv2D(3, (1, 1), padding='same', use_bias=False)(x)  # (3, 8, 8)
    outputs.append(x)
    x = Dropout(0.5)(x)
    outputs.append(x)
    x = Concatenate()([
        MaxPooling2D((2, 2))(x),
        AveragePooling2D((2, 2))(x)])  # (6, 4, 4)
    outputs.append(x)

    x = Flatten()(x)  # (1, 1, 96)
    x = Dense(4, use_bias=False)(x)
    outputs.append(x)
    x = Dense(3)(x)  # (1, 1, 3)
    outputs.append(x)

    outputs.append(Add()([inputs[26], inputs[30], inputs[30]]))
    outputs.append(Subtract()([inputs[26], inputs[30]]))
    outputs.append(Multiply()([inputs[26], inputs[30], inputs[30]]))
    outputs.append(Average()([inputs[26], inputs[30], inputs[30]]))
    outputs.append(Maximum()([inputs[26], inputs[30], inputs[30]]))
    outputs.append(Concatenate()([inputs[26], inputs[30], inputs[30]]))

    # TensorFlow 2.16 no longer puts
    # "inbound_nodes": []
    # for such nested models.
    # todo: Check if the situation resolved with later versions.
    if False:
        intermediate_input_shape = (3,)
        intermediate_in = Input(intermediate_input_shape)
        intermediate_x = intermediate_in
        intermediate_x = Dense(8)(intermediate_x)
        intermediate_x = Dense(5, name='duplicate_layer_name')(intermediate_x)
        intermediate_model = Model(
            inputs=[intermediate_in], outputs=[intermediate_x],
            name='intermediate_model')
        intermediate_model.compile(loss='mse', optimizer='nadam')

        x = intermediate_model(x)[0]  # (1, 1, 5)

        intermediate_model_2 = Sequential(name="intermediate_model_2")
        intermediate_model_2.add(Dense(7, input_shape=(5,)))
        intermediate_model_2.add(Dense(5, name='duplicate_layer_name'))
        intermediate_model_2.compile(optimizer='rmsprop',
                                     loss='categorical_crossentropy')

        x = intermediate_model_2(x)  # (1, 1, 5)

        intermediate_model_3_nested = Sequential(name="intermediate_model_3_nested")
        intermediate_model_3_nested.add(Dense(7, input_shape=(6,)))
        intermediate_model_3_nested.compile(optimizer='rmsprop', loss='categorical_crossentropy')

        intermediate_model_3 = Sequential(name="intermediate_model_3")
        intermediate_model_3.add(Dense(6, input_shape=(5,)))
        intermediate_model_3.add(intermediate_model_3_nested)
        intermediate_model_3.add(Dense(8))
        intermediate_model_3.compile(optimizer='rmsprop', loss='categorical_crossentropy')

        x = intermediate_model_3(x)  # (1, 1, 8)

        x = Dense(3)(x)  # (1, 1, 3)

    shared_activation = Activation('tanh')

    outputs = outputs + [
        Activation('tanh')(inputs[25]),
        Activation('hard_sigmoid')(inputs[25]),
        Activation('selu')(inputs[25]),
        Activation('sigmoid')(inputs[25]),
        Activation('softplus')(inputs[25]),
        Activation('softmax')(inputs[25]),
        Activation('relu')(inputs[25]),
        Activation('relu6')(inputs[25]),
        Activation('swish')(inputs[25]),
        Activation('exponential')(inputs[25]),
        Activation('gelu')(inputs[25]),
        Activation('softsign')(inputs[25]),
        Activation('celu')(inputs[25]),
        Activation('elu')(inputs[25]),
        Activation('exponential')(inputs[25]),
        Activation('gelu')(inputs[25]),
        Activation('hard_tanh')(inputs[25]),
        Activation('linear')(inputs[25]),
        Activation('log_sigmoid')(inputs[25]),
        Activation('silu')(inputs[25]),
        Activation('softmax')(inputs[25]),
        Activation('softsign')(inputs[25]),
        Activation('sparse_plus')(inputs[25]),
        Activation('squareplus')(inputs[25]),
        Activation('tanh')(inputs[25]),
        Activation('tanh_shrink')(inputs[25]),
        Activation('leaky_relu')(inputs[25]),
        LeakyReLU(name="real_LeakyReLU_layer", negative_slope=0.5)(inputs[25]),
        ReLU(name="real_ReLU_layer_1")(inputs[25]),
        ReLU(name="real_ReLU_layer_2", max_value=0.4, negative_slope=1.1, threshold=0.3)(inputs[25]),
        ELU(name="real_ELU_layer")(inputs[25]),
        PReLU(name="real_PReLU_layer_1")(inputs[24]),
        PReLU(name="real_PReLU_layer_2")(inputs[25]),
        PReLU(name="real_PReLU_layer_3")(inputs[26]),
        Softmax(name="real_Softmax_layer")(inputs[25]),
        shared_activation(inputs[25]),
        Activation('linear')(inputs[26]),
        Activation('linear')(inputs[23]),
        activations.celu(inputs[25]),
        activations.elu(inputs[25], alpha=0.71),
        activations.exponential(inputs[25]),
        activations.gelu(inputs[25]),
        activations.hard_shrink(inputs[25], threshold=0.31),
        activations.hard_sigmoid(inputs[25]),
        activations.hard_tanh(inputs[25]),
        activations.leaky_relu(inputs[25], negative_slope=0.31),
        activations.linear(inputs[25]),
        activations.log_sigmoid(inputs[25]),
        activations.log_softmax(inputs[25]),
        activations.relu(inputs[25], negative_slope=0.1, max_value=0.8, threshold=0.3),
        activations.relu6(inputs[25]),
        activations.selu(inputs[25]),
        activations.sigmoid(inputs[25]),
        activations.silu(inputs[25]),
        activations.softmax(inputs[25]),
        activations.soft_shrink(inputs[25], threshold=0.31),
        activations.softplus(inputs[25]),
        activations.softsign(inputs[25]),
        activations.sparse_plus(inputs[25]),
        activations.squareplus(inputs[25], b=3),
        activations.tanh(inputs[25]),
        activations.tanh_shrink(inputs[25]),
        activations.threshold(inputs[25], 0.123, 0.423),
        x,
        shared_activation(x),
    ]

    model = Model(inputs=inputs, outputs=outputs, name='test_model_exhaustive')
    model.compile(loss='mse', optimizer='nadam')

    # fit to dummy data
    training_data_size = 2
    data_in = generate_input_data(training_data_size, input_shapes)
    initial_data_out = model.predict(data_in)
    data_out = generate_output_data(training_data_size, initial_data_out)
    model.fit(data_in, data_out, epochs=10)
    return model


def get_test_model_embedding() -> Model:
    """Returns a minimalistic test model for the Embedding and CategoryEncoding layers."""

    input_dims = [
        1023,  # maximum integer value in input data
        255,
        15,
    ]
    input_shapes: list[tuple[int, ...]] = [
        (100,),  # must be single-element tuple (for sequence length)
        (1000,),
        (1,),
    ]
    assert len(input_dims) == len(input_shapes)
    output_dims = [8, 3]  # embedding dimension

    inputs = [Input(shape=s) for s in input_shapes]

    outputs = []
    for k in range(2):
        embedding = Embedding(input_dim=input_dims[k], output_dim=output_dims[k])(inputs[k])
        outputs.append(embedding)

    outputs.append(CategoryEncoding(1024, output_mode='multi_hot', sparse=False)(inputs[0]))
    # No longer working since TF 2.16: https://github.com/tensorflow/tensorflow/issues/65390
    # Error: Value passed to parameter 'values' has DataType float32 not in list of allowed values: int32, int64
    # outputs.append(CategoryEncoding(1024, output_mode='count', sparse=False)(inputs[0]))
    # outputs.append(CategoryEncoding(16, output_mode='one_hot', sparse=False)(inputs[2]))
    # Error: Value passed to parameter 'values' has DataType float32 not in list of allowed values: int32, int64
    # outputs.append(CategoryEncoding(1023, output_mode='multi_hot', sparse=True)(inputs[0]))

    model = Model(inputs=inputs, outputs=outputs, name='test_model_embedding')
    model.compile(loss='mse', optimizer='adam')

    # fit to dummy data
    training_data_size = 2
    data_in = generate_integer_input_data(training_data_size, 0, input_dims, input_shapes)
    initial_data_out = model.predict(data_in)
    data_out = generate_output_data(training_data_size, initial_data_out)
    model.fit(data_in, data_out, epochs=1)
    return model


def get_test_model_variable() -> Model:
    """Returns a model with variably shaped input tensors."""

    input_shapes = [
        (None, None, 1),
        (None, None, 3),
        (None, 4),
    ]

    inputs = [Input(shape=s) for s in input_shapes]

    outputs = []

    # same as axis=-1
    outputs.append(Concatenate()([inputs[0], inputs[1]]))
    outputs.append(Conv2D(8, (3, 3), padding='same', activation='elu')(inputs[0]))
    outputs.append(Conv2D(8, (3, 3), padding='same', activation='relu')(inputs[1]))
    outputs.append(GlobalMaxPooling2D()(inputs[0]))
    outputs.append(Reshape((2, -1))(inputs[2]))
    outputs.append(Reshape((-1, 2))(inputs[2]))
    outputs.append(MaxPooling2D()(inputs[1]))
    outputs.append(AveragePooling1D(2)(inputs[2]))

    outputs.append(PReLU(shared_axes=[1, 2])(inputs[0]))
    outputs.append(PReLU(shared_axes=[1, 2])(inputs[1]))
    outputs.append(PReLU(shared_axes=[1, 2, 3])(inputs[1]))
    outputs.append(PReLU(shared_axes=[1])(inputs[2]))

    model = Model(inputs=inputs, outputs=outputs, name='test_model_variable')
    model.compile(loss='mse', optimizer='nadam')

    # fit to dummy data
    training_data_size = 2
    data_in = generate_input_data(training_data_size, input_shapes)
    initial_data_out = model.predict(data_in)
    data_out = generate_output_data(training_data_size, initial_data_out)
    model.fit(data_in, data_out, epochs=10)
    return model


def get_test_model_autoencoder() -> Model:
    """Returns a minimal autoencoder test model."""
    input_img = Input(shape=(1,), name='input_img')
    encoded = Identity()(input_img)  # Since it's about testing node connections, this suffices.
    encoder = Model(input_img, encoded, name="encoder")

    input_encoded = Input(shape=(1,), name='input_encoded')
    decoded = Identity()(input_encoded)
    decoder = Model(input_encoded, decoded, name="decoder")

    autoencoder_input = Input(shape=(1,), name='input_autoencoder')
    x = encoder(autoencoder_input)
    autoencodedanddecoded = decoder(x)
    autoencoder = Model(inputs=autoencoder_input, outputs=autoencodedanddecoded, name="autoencoder")
    autoencoder.compile(optimizer='sgd', loss='mse')
    return autoencoder


def get_test_model_sequential() -> Model:
    """Returns a typical (VGG-like) sequential test model."""
    model = Sequential()
    model.add(Conv2D(8, (3, 3), activation='relu', input_shape=(32, 32, 3)))
    model.add(Conv2D(8, (3, 3), activation='relu'))
    model.add(Permute((3, 1, 2)))
    model.add(MaxPooling2D(pool_size=(2, 2)))
    model.add(Permute((2, 3, 1)))
    model.add(Dropout(0.25))

    model.add(Conv2D(16, (3, 3), activation='elu'))
    model.add(Conv2D(16, (3, 3)))
    model.add(ELU())

    model.add(MaxPooling2D(pool_size=(2, 2)))
    model.add(Dropout(0.25))

    model.add(Flatten())
    model.add(Dense(64, activation='sigmoid'))
    model.add(Dropout(0.5))
    model.add(Dense(10, activation='softmax'))

    model.compile(loss='categorical_crossentropy', optimizer='sgd')

    # fit to dummy data
    training_data_size = 2
    data_in = [np.random.random(size=(training_data_size, 32, 32, 3))]
    data_out = [np.random.random(size=(training_data_size, 10))]
    model.fit(data_in, data_out, epochs=10)
    return model


def main() -> None:
    """Generate different test models and save them to the given directory."""
    if len(sys.argv) != 3:
        print('usage: [model name] [destination file path]', flush=True)
        sys.exit(1)
    else:
        model_name = sys.argv[1]
        dest_path = sys.argv[2]

        get_model_functions = {
            'exhaustive': get_test_model_exhaustive,
            'embedding': get_test_model_embedding,
            'variable': get_test_model_variable,
            'autoencoder': get_test_model_autoencoder,
            'sequential': get_test_model_sequential,
        }

        if not model_name in get_model_functions:
            print('unknown model name: ', model_name)
            sys.exit(2)

        np.random.seed(0)

        model_func = get_model_functions[model_name]
        model = model_func()
        model.save(dest_path)

        # Make sure models can be loaded again,
        # see https://github.com/fchollet/keras/issues/7682
        model = load_model(dest_path)
        model.summary()
        # plot_model(model, to_file= str(model_name) + '.png', show_shapes=True, show_layer_names=True)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import numpy as np
from keras.layers import Input, Dense
from keras.models import Model

inputs = Input(shape=(4,))
x = Dense(5, activation='relu')(inputs)
predictions = Dense(3, activation='softmax')(x)
model = Model(inputs=inputs, outputs=predictions)
model.compile(loss='categorical_crossentropy', optimizer='nadam')

model.fit(
    np.asarray([[1, 2, 3, 4], [2, 3, 4, 5]]),
    np.asarray([[1, 0, 0], [0, 0, 1]]), epochs=10)

model.save('readme_example_model.keras')
